/*************************************************************************
 *
 * BumsBox 1.0
 *
 * Controller : PIC 16F76
 * Oszillator : Quarz 10 Mhz
 * Editor : MI-Edit, Tab-Space 2
 *
 * (C) Fingers elektrische Welt 2005
 *
 ************************************************************************/

#include	<pic.h>
#include	<stdlib.h>
#include	<string.h>
#include	<stdio.h>

#include  "defines.h"                                // Konstantendeklarationen

__CONFIG(PWRTEN & WDTEN & HS & BORDIS);

// Timerfunktionen
struct Timestruct
 {
   volatile char sekunde;		                         // zhlt Sekunden
   volatile unsigned long Universaltimer;            // Wird im 20ms-Takt hochgezhlt
   volatile Ticker;                                  // Zur Synchronisation
 } Zeit;

// Puffervariablen
char outpbuf [20];                                   // Puffer zur Wandlung Int -> ASCII

// Steuerung der Zndspannungserzeugung
volatile char ZuendFlag = FALSE;

// Gemessene Spannungen
volatile char Zuendkreisspannung =0;
volatile char Batteriespannung = 0;
volatile char Zwischenkreisspannung = 0;

// Zu steuernder Znder
char Zuendernummer = 0;

// Empfangenes Kommando
char Kommando = KeinKommando;

/***********************************************************************************
 *
 * Schaltet SDA und SCL als Eingang und macht die Schnittstelle startklar
 *
 ***********************************************************************************/
void i2c_init (void)
 {

  TRISC &= ~(8+16);
  SDA = 1;
  SCL = 1;

 }

/**********************
 *
 * Delay ms
 *
 **********************/
void Delayms (int time)
 {
  int i;

  for (i=0; i < time; i++) DelayUs (1000);
 }

/*******************************************
 *
 * Schaltet die Zndspannung an den Ausgang
 *
 ********************************+*********/
void Scharf (char wert)
 {
  if (wert) PORTB |= 16; else PORTB &= ~16;
 }

/*************************************************************************************
 *
 * Schreibt ein Byte via i2C, erzeugt kein Start/Stop, wertet ACK vom Slave nicht aus
 *
 *************************************************************************************/
void i2c_writebyte (unsigned char byte)
 {
  unsigned char i;

  for (i=0; i<8; i++)
   {
     if (byte>127) SDA=1; else SDA=0;                           // SDA bedienen
     DelayUs (uDLY);                                            // kurze Pause
     SCL=1; DelayUs (uDLY);                                     // SCL = 1 + Pause
     SCL=0; DelayUs (uDLY);                                     // SCL = 0 + Pause
     byte <<= 1;
    }

  SDA=1; DelayUs (uDLY);                                        // SDA = 1 + Pause
  SCL=1; DelayUs (uDLY);                                        // SCL = 1 + Pause

  // 9. Bit erzeugen, damit der Slave das Acknowledge loswerden kann
  TRISC |= SDA_PORT;                                            // SDA als Eingang schalten
  SCL = 0; DelayUs (uDLY);                                      // SCL = 0 + Pause um ACK abzuwarten
  TRISC &= ~SDA_PORT;                                           // SDA als Ausgang schalten

 }

/***********************************************************************************
 *
 * Liest ein Byte via i2C, erzeugt bei Bedarf ein Acknowledge
 *
 ***********************************************************************************/
char i2c_readbyte (char Acknowledge)
 {
  char i, byte=0;

  TRISC |= SDA_PORT;                                              // SDA als Eingang schalten

  for (i=0; i<8; i++)                                             // Bits rausschieben
   {
     byte <<= 1;
     if (SDA) byte |= 1;
     DelayUs (rDLY);
     SCL=1; DelayUs (rDLY);
     SCL=0; DelayUs (rDLY);
   }

  // 9. Bit erzeugen
  TRISC &= ~SDA_PORT;                                              // SDA als Ausgang schalten

  if (Acknowledge) SDA = 0; else SDA = 1;
  DelayUs (rDLY);                                                  // Pause
  SCL=1; DelayUs (rDLY);                                           // SCL = 1 + Pause
  SCL=0; DelayUs (rDLY);                                           // SCL = 0 + Pause
  SDA=1;DelayUs (rDLY);                                            // SDA = 1 + Pause

  return (byte);
}

/***********************************************************************************
 *
 * Schreibt ein Byte via i2C an Adresse "adress", erzeugt Start/Stop
 *
 ***********************************************************************************/
void i2c_write (char deviceaddr, char data)
 {
    char byte;

    // Start-Condition erzeugen
    SDA=1; SCL=1; DelayUs (uDLY);                              // Beide Leitungen auf 1
    SDA=0; DelayUs (uDLY);                                     // SDA = 0 + Pause
    SCL=0; DelayUs (uDLY);                                     // SCL = 0 + Pause

    // Slave adressieren
    byte = deviceaddr;                                         // Slaveadresse zuweisen
    i2c_writebyte (byte);                                      // Slaveadresse schreiben

    // Datum senden
    byte = data;                                               // Datum setzen
    i2c_writebyte (byte);                                      // Byte senden

    // Stop-Condition erzeugen
    SDA=0; DelayUs (uDLY);                                     // SDA = 0 + Pause
    SCL=1; DelayUs (uDLY);                                     // SCL = 1 + Pause
    SDA=1; DelayUs (uDLY);                                     // SDA = 1 + Pause
 }

/***********************************************************************************
 *
 * Liest ein Byte via i2C, erzeugt Start/Stop
 *
 ***********************************************************************************/
char i2c_read (unsigned char deviceaddr)
 {
  char byte;

  // Start-Condition erzeugen
  SCL=1; DelayUs (rDLY);                                        // SCL = 1 + Pause
  SDA=0; DelayUs (rDLY);                                        // SDA = 0 + Pause
  SCL=0; DelayUs (rDLY);                                        // SCL = 0 + Pause

  // Slave zum Lesen adressieren
  byte = deviceaddr+1;                                          // Leseadresse setzen
  i2c_writebyte (byte);                                         // Adresse senden

  byte = i2c_readbyte (FALSE);                                  // Ein Byte lesen

  // Stop-Condition erzeugen
  SDA=0; DelayUs (rDLY);                                        // SDA = 0 + Pause
  SCL=1; DelayUs (rDLY);                                        // SCL = 1 + Pause
  SDA=1; DelayUs (rDLY);                                        // SCL = 1 + Pause

  return (byte);
 }

/**********************
 *
 * Initialisierung
 *
 **********************/
void init (void)
 {

   // IO-Konfiguration (1 = Input)
   TRISA = 0b11111111;		                            // Nur Wandlerkanle
   TRISB = 0b11000000;                                // Nur Ausgnge
   TRISC = 0b11100000;                                //
   PORTB = 0;
   PORTC = 0;

   Scharf (FALSE);                                    // Zndspannung abschalten

   // Timer klarmachen
   TMR1IE=1;				                                  // Timer1 overflow Int enabled
   T1CON = 0x01;
   INTCON = 0xc0;				                              // Global Interrupts enabled

   // Serielle Schnittstelle
   SPBRG=0x0f;				                                // 9600 Baud (bei BRGH=0), Normalbetrieb
   TXSTA=0x20;				                                // 8-Bit, Transmit enabled, syncr. Mode, low Speed
   RCSTA=0x90;				                                // serial Port enabled, 8-Bit

   // Einstellen des AD-Wandlers
   ADCON1 = 0x16;				                              // keine AD-Wandlerfunktion eingestellt

   // I2C-Krempel an den Start rollen-------
   i2c_init ();                                       // I2C-Klarmachen
   i2c_write (Output0, 0);                            // Ausgnge 0..7 abschalten
   i2c_write (Output1, 0);                            // Ausgnge 8..15 abschalten

 }

/**********************************************
 *
 * Text via serieller Schnittstelle rauspumpen
 *
 **********************************************/
void print (const char *data)
 {
  char cddummy;

  if (FERR) cddummy = RCREG;
  if (OERR)
  	{
    	CREN=0;
    	CREN=1;
  	}
  	do
  	 {
  		while (!TRMT);
  		Delayms (20);
    	TXREG = *data;
  	 } while ( *++data );

 }


/*********************************************************************
 *
 * Lesen, ob ein Byte ber die serielle Schnittstelle gekommen sind.
 * Bei Fehler oder keinen Daten wird 0xff zurckgegeben.
 *
 ********************************************************************/
char read_byte_usart (void)
 {
  unsigned char tempvar;

  if (RCIF==1)
  	{
  	 if ( (RCSTA & 0x06) != 0)		 // Fehler aufgetreten ?
  		{
  		  tempvar=RCREG;		         // Ja, Daten wegwerfen,
        RCIF = 0;                     // Flag rcksetzen
  		  return 0xff;			         // 0xff zurckliefern
  		} else                       // From  if ( (RCSTA & 0x06) != 0)
  		{

  		 tempvar=RCREG;
       RCIF = 0;                     // Flag rcksetzen
   		 return (tempvar);		         // Daten zurckliefern
   	  }                            // From else if ( (RCSTA & 0x06) != 0)

  	 }                             // From if (RCIF==1)
  else return 0xff; 				       // Keine Daten da
 }

/***********************************************************
 *
 * Analog Port einlesen
 *
 **********************************************************/
char GetAnalog (char Port)
 {
   ADCON0 = 0b10000001;                    // ADC ein
   if (Port == 0) ADCON0 |= 0b00000000;    // Kanal whlen
   if (Port == 1) ADCON0 |= 0b00001000;    // Kanal whlen
   if (Port == 2) ADCON0 |= 0b00010000;    // Kanal whlen

   Delayms (10);

   ADCON1=0x00;		 	                       // V-Ref = Betriebsspannung alle Analogports ein
   ADGO=1;		                             // Wandlung starten
   while (ADGO!=0); 	                     // warten bis Wandlung fertig

   return ADRES;
 }

/********************************************
 *
 * Liefert den Rohwert der Batteriespannung
 *
 ********************************************/
char GetBatterieRaw (void)
 {
  return (GetAnalog (1));
 }


/*************************************************
 *
 * Liefert den die Batteriespannung (*10) in Volt
 *
 *************************************************/
char GetBatterie (void)
 {
  int temp;

  temp = GetBatterieRaw ();
  temp *= 50;
  temp /= 255;

  return ((char) temp);
 }

/************************************************
 *
 * Liefert den Rohwert der Zwischenkreisspannung
 *
 ************************************************/
char GetDCDCRaw (void)
 {
  return (GetAnalog (2));
 }

/************************************************
 *
 * Liefert den die Zwischenkreisspannung in Volt
 *
 ************************************************/
char GetDCDC (void)
 {
  int temp;

  temp = GetDCDCRaw ();
  temp *= 16;
  temp /= 255;

  return ((char) temp);
 }

/***************************************
 *
 * Liefert den Rohwert der Zndspannung
 *
 ***************************************/
char GetZuenderRaw (void)
 {
  return (GetAnalog (0));
 }

/***********************************
 *
 * Liefert die Zndspannung in Volt
 *
 ***********************************/
char GetZuendspannung (void)
 {
  int temp;

  temp = GetZuenderRaw ();
  temp *= 48;
  temp /= 255;

  return ((char) temp);
 }

/*************************************
 *
 * Schaltet die Zwischenkreisspannung
 *
 *************************************/
void DCDC (char wert)
 {
  if (wert) PORTB |= 4; else PORTB &= ~4;
 }

/******************************
 *
 * Entldt den Zndkondensator
 *
 ******************************/
void Discharge (char wert)
 {
  if (wert) PORTB |= 1; else PORTB &= ~1;
 }

/********************************
 *
 * Pumpt den Zndkondensator auf
 *
 ********************************/
void Boost (char wert)
 {

  // Timer 2 setzen
  PR2 = 0xFF;

  // Puls/Pausenverhltniss setzen
  CCPR1L = 0x3F;
  CCP1CON |= 16;

  // CCP1 = Output
  TRISC &= ~4;

  // Timer 2 anschmeissen
  T2CON = 0b00000000;
  TMR2ON = 1;

  // PWM ein
  if (wert) CCP1CON |= 0b00001100; else CCP1CON = 0;

 }

/**********************************************************************
 *
 * Timer 1 Overflow-ISR und RCI-ISR (Receive complete) und SPI
 *
 **********************************************************************/
void interrupt ISR (void)
 {

    // ------------- Timer 1-Overflow-Interrupt aufgelaufen ? -------------
    if (TMR1IF == 1)
     {
        Zeit.Universaltimer++;
        Zeit.Ticker++;

    		TMR1ON=0;		                                     // Timer 1 Stop
    		TMR1L=0;		                                     // Timer 1L laden
    		TMR1H=68;		                                     // Timer 1H laden
    							                                       // mit 68 --> 20ms bis zum overflow
    		TMR1ON=1;		                                     // Timer 1 Start
    		TMR1IF = 0;                                      // Interrupt-Flag zurcksetzen

        // LED blinken lassen
        if (Zeit.Ticker == 20)
         {
          Zeit.Ticker = 0;
          if (PORTB & 32) PORTB &= ~32; else PORTB |= 32;
         } // Von if (Zeit.Ticker == 20)

    	}

    // ------------- UART-Receive-Interrupt aufgelaufen ? -------------------------

    /*
    if (RCIF == 1)
     {
      RCIF = 0;
 		  tempvar = RCREG;		                                 // Daten wegwerfen,
    	}
    */
    // ------------- UART-Transmit-Interrupt aufgelaufen ? -------------------------

    if (TXIF == 1)
     {
      TXIF = 0;
     }

  return;

 }

/*************************
 *
 * Messung der Spannungen
 *
 *************************/
void Spannungsmessung (void)
 {
  Zuendkreisspannung = GetZuendspannung ();
  Delayms (10);
  Batteriespannung = GetBatterie ();
  Delayms (10);
  Zwischenkreisspannung = GetDCDC ();
 }


/**************************************************
 *
 * Interpretiert via RS232 reingekommene Kommando
 *
 **************************************************/
char KommandoInterpreter (void)
 {
  #define Puflen 8														// Gre des Puffers
  #define Timeout	100													// 200ms Timeout

  char Zeichen;																// Empfangenes Zeichen
  char Zeiger = 0;														// zeiger auf Puffer
  char Kommandopuffer [Puflen];	  						// Zwischenspeicher fr Kommandos
  unsigned long Ticker;												// Timeoutzhler

  Zeichen = read_byte_usart ();
  if (Zeichen == 0xFF) return (KeinKommando);

  Kommandopuffer [0] = Zeichen;
  Zeiger = 1;

  //---------------------------------------------------
  // Zeichen empfangen bis Timeout, berlauf oder Ende
  //---------------------------------------------------
  Ticker = Zeit.Universaltimer;
  while ( (Zeiger < Puflen) && (Zeichen != CR) && ( Zeit.Universaltimer < (Ticker+Timeout) ) )
   {
    Zeichen = read_byte_usart ();
    if (Zeichen != 0xff)
     {
      Kommandopuffer [Zeiger] = Zeichen;
      Zeiger++;
     } // Von if (Zeichen != 0)
   } // Von while ( (Zeiger < 10) && (Zeichen != CR) && () )

  // Ist ein Timeout aufgetreten ?
  if (Zeit.Universaltimer >= (Ticker+Timeout)) return (KeinKommando);

  // Ist ein berlauf aufgetreten ?
  if (Zeiger >= Puflen) return (KeinKommando);

  //-----------------------
  // Ein Kommando ist drin
  //-----------------------

  Kommandopuffer [Zeiger] = '\0';
  // print (Kommandopuffer);

  // Spannungen messen
  if (strstr (Kommandopuffer, "MESS") != NULL) return (KommandoSpannungMessen);

  // Zndspannung einschalten
  if (strstr (Kommandopuffer, "ZEIN") != NULL) return (KommandoSpannungEin);

  // Zndspannung abschalten
  if (strstr (Kommandopuffer, "ZAUS") != NULL) return (KommandoSpannungAus);

  // Zndspannung abbauen
  if (strstr (Kommandopuffer, "DISC") != NULL) return (KommandoSpannungAbbauen);

  // Znder einschalten (Fxx)
  if (strstr (Kommandopuffer, "F") != NULL)
   {
    Kommandopuffer [0] = ' ';
    Kommandopuffer [3] = '\0';
    Zuendernummer = (char)atoi (Kommandopuffer);
    Zuendernummer -= 1;
    if (Zuendernummer < 16)  return (KommandoZuenden); else
                             return (KeinKommando);
   }

  //----------------------------------------------------
  // Hier gelandet konnte es nicht interpretiert werden
  //----------------------------------------------------
  return (KeinKommando);

 }

/*************************************************************************
 *
 * Hauptprogramm
 *
 *************************************************************************/
main (void)
{

 init ();

 /*
 sprintf (outpbuf, "START\r");
 print (outpbuf);
 */

  while (1)
   {

    CLRWDT ();

    Spannungsmessung ();

    //---------------------------
    // Regelung der Zndspannung
    //---------------------------

    if (ZuendFlag)
     {

      // Spannung am Zndkreis regeln
      if (Zuendkreisspannung < 37) Boost (TRUE); else Boost (FALSE);

      // berprfung auf vorhandene Zwischenkreisspannung
      if (Zwischenkreisspannung < 10) CCPR1L = 0x1F; // Puls/Pausenverhltniss ndern

      // berprfung auf kritische  Zwischenkreisspannung
      if (Zwischenkreisspannung < 5)
       {
         Boost (FALSE);       // PWM aus
         TRISC |= 4;          // Pin hochohmig schalten
         Delayms (300);       // Kurze Pause
       }

      } // von if (ZuendFlag)



    //------------------------------------------------
    // Interpretation eventuell reinkommender Befehle
    //------------------------------------------------

    Kommando = KommandoInterpreter ();

    // Spannung messen
    if (Kommando == KommandoSpannungMessen)
     {
      sprintf (outpbuf, "%2d%2d%2d\r", Batteriespannung, Zwischenkreisspannung, Zuendkreisspannung );

      print (outpbuf);
     }

    // Zndspannung einschalten
    if (Kommando == KommandoSpannungEin)
     {
      print ("ZON\r");
      Discharge (FALSE);                    // Entladung abschalten
      DCDC (TRUE);                          // Zwischenkreis an
      Delayms (300);                        // Kurze Pause
      ZuendFlag = TRUE;                     // Flag setzen
     }

    // Zndspannung abschalten
    if (Kommando == KommandoSpannungAus)
     {
      print ("ZOFF\r");
      DCDC (FALSE);                         // Zwischenkreis aus
      Boost (FALSE);                        // Booster aus
      TRISC |= 4;                           // Pin hochohmig schalten
      Discharge (FALSE);                    // Entladen aus
      ZuendFlag = FALSE;                    // Flag setzen
     }

    // Zndspannung abbauen
    if (Kommando == KommandoSpannungAbbauen)
     {
      print ("DON\r");
      ZuendFlag = FALSE;                    // Flag setzen
      DCDC (FALSE);                         // Zwischenkreis aus
      Boost (FALSE);                        // Booster aus
      TRISC |= 4;                           // Pin hochohmig schalten
      Discharge (TRUE);                     // Entladen
     }

    // Znden
    if (Kommando == KommandoZuenden)
     {

      DCDC (FALSE);                                      // Zwischenkreis aus
      Boost (FALSE);                                     // Booster aus
      TRISC |= 4;                                        // Pin hochohmig schalten
      Scharf (TRUE);                                     // Spannung an den Ausgang schalten
      Delayms (200);                                     // Warten, bis Relais angezogen ist

      //-------------------------------------------
      // sprintf (outpbuf, " %d \n\r", Zuendernummer);
      // print (outpbuf);
      //-------------------------------------------

      if (Zuendernummer < 8)
       { // Ersten Portexpander ansprechen
        i2c_write (Output0, (1 << Zuendernummer));        // Ausgang einschalten
        Delayms (800);                                    // Kurze Pause
        i2c_write (Output0, 0);                           // Ausgang abschalten
       } else
       { // Zweiten Portexpander ansprechen
        i2c_write (Output1, (1 << (Zuendernummer-8)));    // Ausgang einschalten
        Delayms (800);                                    // Kurze Pause
        i2c_write (Output1, 0);                           // Ausgang abschalten
       }

      Scharf (FALSE);                                     // Spannung am Ausgang abschalten
      if (ZuendFlag) DCDC (TRUE);                         // Zwischenkreis ggf. wieder einschalten

      print ("FOK\r");
     }

    //-----------------------
    // Ende des Interpreters
    //-----------------------

   }	// Von While (1)

}

// Thats all folks.....

